//================================================================================================
// $Id: cgfile.cpp 7226 2024-01-11 19:24:54Z aron $
//
// Project:
//		Welltype Compiler
// Contents:
//		Code generator file helper
// Authors:
//		Aron Barath, 2014-2023
//================================================================================================
// Copyright (c) Aron Barath
//================================================================================================

#include "cgfile.h"
#include "codegen.h"
#include "parserext.h"
#include <cstdio>
#include <cerrno>
#include <cstring>
#include <sstream>

#ifdef _MSC_VER
#  pragma warning ( disable : 4996 )
#endif // _MSC_VER

#define __SWAP_16(_short_) ((wt_u16)( \
		(((wt_u16)(_short_))>>8) | \
		(((wt_u16)(_short_))<<8)  \
	))

#define __SWAP_32(_int_) ((wt_u32)( \
		(((wt_u32)(_int_))<<24) | \
		((((wt_u32)(_int_))<<8) & 0x00ff0000ul) | \
		((((wt_u32)(_int_))>>8) & 0x0000ff00ul) | \
		(((wt_u32)(_int_))>>24) \
	))

#define __SWAP_64(_long_) ((wt_u64)( \
		(((wt_u64)(_long_))<<(8*7)) | \
		((((wt_u64)(_long_))<<(8*5)) & 0x00ff000000000000ull) | \
		((((wt_u64)(_long_))<<(8*3)) & 0x0000ff0000000000ull) | \
		((((wt_u64)(_long_))<<(8*1)) & 0x000000ff00000000ull) | \
		((((wt_u64)(_long_))>>(8*1)) & 0x00000000ff000000ull) | \
		((((wt_u64)(_long_))>>(8*3)) & 0x0000000000ff0000ull) | \
		((((wt_u64)(_long_))>>(8*5)) & 0x000000000000ff00ull) | \
		(((wt_u64)(_long_))>>(8*7)) \
	))

namespace wtc
{
namespace codegen
{

File::File(wtc::ProblemHandler & probhan, const std::string & filename,
	bool swap_byteorder)
	: mProbHan(probhan)
	, mFilename(filename)
	, mSwapByteorder(swap_byteorder)
	, mHeaderOffset(0)
{
	// ok
}

File::~File()
{
	PostProcess();

	FILE* const file = fopen(mFilename.c_str(), "wb");

	if(!file)
	{
		const int err = errno;

		std::ostringstream oss;
		oss << "Could not open for write: " << strerror(err);

		mProbHan.AddError(mFilename, 0, oss.str());
		return;
	}

	if(1!=fwrite(&mData[0], mData.size(), 1, file))
	{
		const int err = errno;

		std::ostringstream oss;
		oss << "Failed to write: " << strerror(err);

		mProbHan.AddError(mFilename, 0, oss.str());
	}

	fclose(file);
}

void File::Write(const void* data, wt_u32 size)
{
	const wt_u8* it  = reinterpret_cast<const wt_u8*>(data);
	const wt_u8* end = it + size;

	while(it<end)
	{
		mData.push_back(*it++);
	}
}

void File::Write(wt_u8 v)
{
	mData.push_back(v);
}

void File::Write(wt_u16 v)
{
	if(mSwapByteorder) { v = __SWAP_16(v); }
	Write(&v, sizeof(v));
}

void File::Write(wt_u32 v)
{
	if(mSwapByteorder) { v = __SWAP_32(v); }
	Write(&v, sizeof(v));
}

void File::Write(wt_u64 v)
{
	if(mSwapByteorder) { v = __SWAP_64(v); }
	Write(&v, sizeof(v));
}

void File::Write(const wtb3_hdr & hdr)
{
	mHeaderOffset = mData.size();
	
	for(unsigned int i=0;i<sizeof(hdr.b_ident);++i)
	{
		Write(hdr.b_ident[i]);
	}

	Write(hdr.b_size);
	Write(hdr.b_shnum);
	Write(hdr.b_thnum);
	Write(hdr.b_shoff);
	Write(hdr.b_thoff);
	Write(hdr.b_ctor);
	Write(hdr.b_dtor);
}

void File::Write(const wtb3_shdr & shdr)
{
	Write(shdr.sh_type);
	Write(shdr.sh_flag);
	Write(shdr.sh_link);
	Write(shdr.sh_offs);
	Write(shdr.sh_size);
	Write(shdr.sh_addr);
}

void File::Write(const wtb3_thdr & thdr)
{
	Write(thdr.th_type);
	Write(thdr.th_flag);
	Write(thdr.th_cssize);
	Write(thdr.th_dssize);
	Write(thdr.th_pcbase);
}

void File::Write(const wtb3_metahdr & mhdr)
{
	Write(mhdr.metacount);
	Write(mhdr.reserved);
}

void File::Write(const wtb3_meta & meta)
{
	Write(meta.meta_key_offs);
	Write(meta.meta_val_offs);
}

void File::Write(const wtb3_symhdr & symhdr)
{
	Write(symhdr.symcount);
	Write(symhdr.reserved);
}

void File::Write(const wtb3_sym & sym)
{
	Write(sym.sym_type);
	Write(sym.sym_name);
	Write(sym.sym_offs);
	Write(sym.sym_info);
}

void File::Write(const wtb3_dlabelhdr & dlabhdr)
{
	Write(dlabhdr.labelcount);
	Write(dlabhdr.reserved);
}

void File::Write(const wtb3_dlabel & dlab)
{
	Write(dlab.addr);
	Write(dlab.lab_offs);
}

void File::Write(const wtb3_integhdr & ihdr)
{
	Write(ihdr.algorithm);
	Write(ihdr.reserved);
}

void File::Write(const std::string & str_zero)
{
	Write(str_zero.c_str(), str_zero.length()+1);
}

void File::AddPostData(PostDataGenerator* gen, wt_u32 length)
{
	mPostData.push_back(PostData(mData.size(), length, gen));

	for(wt_u32 i=0;i<length;++i)
	{
		mData.push_back(0);
	}
}

void File::PostProcess()
{
	for(std::vector<PostData>::iterator it=mPostData.begin(),end=mPostData.end();it!=end;++it)
	{
		std::vector<wt_u8> out;

		it->generator->Generate(out, mData);

		if(out.size() > it->length)
		{
			mProbHan.AddError(mFilename, 0,
				"Invalid data generated during the post process step.");
		}
		else
		{
			std::copy(out.begin(), out.end(), mData.begin()+it->offset);
		}
	}
}

} // codegen
} // wtc
