//================================================================================================
// $Id: cginstrument.cpp 6514 2023-04-03 18:02:07Z aron $
//
// Project:
//		Welltype Compiler
// Contents:
//		Code generator instrument
// Authors:
//		Aron Barath, 2014-2023
//================================================================================================
// Copyright (c) Aron Barath
//================================================================================================

#include "cginstrument.h"
#include "cgsection.h"
#include "codegen.h"
#include "wtparser/ast/ast_functionsignature.h"
#include "wtparser/builtins.h"
#include "welltype/stdlib.h"
#include <sstream>

namespace wtc
{
namespace codegen
{

InstrumentEngine::InstrumentEngine(InstrumentProblemHandler* probhan)
	: mLastGlobal("**")
	, mNextLabelId(1)
	, mProblemHandler(probhan)
{
	for(wtparser::builtin::Builtins::const_iterator it=wtparser::builtin::builtins.begin(),end=wtparser::builtin::builtins.end();it!=end;++it)
	{
		const wtparser::builtin::Builtin & bf = *it;
		instruments_t ins;

		for(wtparser::builtin::InstructionVec::const_iterator iit=bf.GetInstrVec().begin(),iend=bf.GetInstrVec().end();iit!=iend;++iit)
		{
			ins.push_back(Instrument(Instrument::IK_SINGLE, WT_TOMB_GETOPCODE(*iit), WT_TOMB_GETOPDATA(*iit)));
		}

		mEXEC_Table.insert(std::make_pair(bf.GetUniqueId(), ins));
	}
}

std::string InstrumentEngine::GenLabel()
{
	std::ostringstream ss;
	ss << mLastGlobal << '.' << (mNextLabelId++);
	return ss.str();
}

std::string InstrumentEngine::TruncateCanonical(const std::string & name)
{
	std::size_t n = name.find(':');

	if(name.npos!=n)
	{
		while(n>0 && ' '==name[n-1]) { --n; }

		return name.substr(0, n);
	}

	return name;
}

bool InstrumentEngine::ReplaceLast(const wt_u8 opcode_to_replace, wt_u8 new_opcode)
{
	Instrument & last = mInstruments.back();

	if(opcode_to_replace==last.opcode)
	{
		last.opcode = new_opcode;
		return true;
	}

	return false;
}

bool InstrumentEngine::LastIsWith(const wt_u8 opcode) const
{
	return opcode==mInstruments.back().opcode;
}

void InstrumentEngine::MakePushImm(wt_u32 value)
{
	if((value & 0xFFFFFF00u) == 0)
	{
		AddSingle(WT_TOMB_OP_PWZ_I, wt_u8(value));
	}
	else
	if((value & 0xFFFFFF00u) == 0xFFFFFF00u)
	{
		AddSingle(WT_TOMB_OP_PWO_I, wt_u8(value));
	}
	else
	if((value & 0xFF000000u) == 0)
	{
		AddDouble(WT_TOMB_OP_PWZ_L, wt_u8(value>>16), wt_u16(value));
	}
	else
	if((value & 0xFF000000u) == 0xFF000000u)
	{
		AddDouble(WT_TOMB_OP_PWO_L, wt_u8(value>>16), wt_u16(value));
	}
	else
	{
		AddDouble(WT_TOMB_OP_PWZ_L, 0, wt_u16(value>>16));

		if(value & 0x0000ff00u)
		{
			AddDouble(WT_TOMB_OP_LSOR_L, 0, wt_u16(value));
		}
		else
		{
			AddSingle(WT_TOMB_OP_LSOR_I, wt_u8(value));
		}
	}
}

void InstrumentEngine::MakePushImm(wt_u64 value)
{
	if((value & 0xFFFFFFFFFFFFFF00ul) == 0)
	{
		AddSingle(WT_TOMB_OP_PWZL_I, wt_u8(value));
	}
	else
	if((value & 0xFFFFFFFFFFFFFF00ul) == 0xFFFFFFFFFFFFFF00ul)
	{
		AddSingle(WT_TOMB_OP_PWOL_I, wt_u8(value));
	}
	else
	if((value & 0xFFFFFFFFFF000000ul) == 0)
	{
		AddDouble(WT_TOMB_OP_PWZL_L, wt_u8(value>>16), wt_u16(value));
	}
	else
	if((value & 0xFFFFFFFFFF000000ul) == 0xFFFFFFFFFF000000ul)
	{
		AddDouble(WT_TOMB_OP_PWOL_L, wt_u8(value>>16), wt_u16(value));
	}
	else
	if((value & 0xFFFFFFFF00000000ul) == 0)
	{
		AddSingle(WT_TOMB_OP_PWZL_I, wt_u8(value>>24));
		AddDouble(WT_TOMB_OP_LSORL_L, wt_u8(value>>16), wt_u16(value));
	}
	else
	if((value & 0xFFFFFFFF00000000ul) == 0xFFFFFFFF00000000ul)
	{
		AddSingle(WT_TOMB_OP_PWOL_I, wt_u8(value>>24));
		AddDouble(WT_TOMB_OP_LSORL_L, wt_u8(value>>16), wt_u16(value));
	}
	else
	if((value & 0xFFFF000000000000ul) == 0)
	{
		AddDouble(WT_TOMB_OP_PWZL_L, wt_u8(value>>40), wt_u16(value>>24));
		AddDouble(WT_TOMB_OP_LSORL_L, wt_u8(value>>16), wt_u16(value));
	}
	else
	if((value & 0xFFFF000000000000ul) == 0xFFFF000000000000ul)
	{
		AddDouble(WT_TOMB_OP_PWOL_L, wt_u8(value>>40), wt_u16(value>>24));
		AddDouble(WT_TOMB_OP_LSORL_L, wt_u8(value>>16), wt_u16(value));
	}
	else
	{
		wt_u32 up = 0x0000FFFFu & wt_u32(value >> (24+24));
		wt_u32 mi = 0x00FFFFFFu & wt_u32(value >> (24));
		wt_u32 lo = 0x00FFFFFFu & wt_u32(value);
		AddDouble(WT_TOMB_OP_PWZL_L, wt_u8(up>>16), wt_u16(up));
		AddDouble(WT_TOMB_OP_LSORL_L, wt_u8(mi>>16), wt_u16(mi));
		AddDouble(WT_TOMB_OP_LSORL_L, wt_u8(lo>>16), wt_u16(lo));
	}
}

void InstrumentEngine::MakeST(wt_u32 size)
{
	switch(size)
	{
	case 1:
		AddSingle(WT_TOMB_OP_STB, 0);
		break;

	case 2:
		AddSingle(WT_TOMB_OP_STH, 0);
		break;

	default:
		Make_Generic(size, WT_TOMB_OP_ST_I, WT_TOMB_OP_ST_L, WT_TOMB_OP_ST_S);
		break;
	}
}

void InstrumentEngine::MakeLD(wt_u32 size)
{
	switch(size)
	{
	case 1:
		AddSingle(WT_TOMB_OP_LDB, 0);
		break;

	case 2:
		AddSingle(WT_TOMB_OP_LDH, 0);
		break;

	default:
		Make_Generic(size, WT_TOMB_OP_LD_I, WT_TOMB_OP_LD_L, WT_TOMB_OP_LD_S);
		break;
	}
}

void InstrumentEngine::Make_Generic(wt_u32 val, wt_u8 op_i, wt_u8 op_l)
{
	if(val & 0xFF000000u)
	{
		mProblemHandler->InstrumentError("Value too large for intrument.");
	}
	else
	if(val & 0x00FFFF00u)
	{
		AddDouble(op_l, wt_u8(val>>16), wt_u16(val));
	}
	else
	{
		AddSingle(op_i, wt_u8(val));
	}
}

void InstrumentEngine::Make_Generic(wt_u32 val, wt_u8 op_i, wt_u8 op_l, wt_u8 op_s)
{
	if(val & 0xFF000000u)
	{
		MakePushImm(val);
		AddSingle(op_s);
	}
	else
	if(val & 0x00FFFF00u)
	{
		AddDouble(op_l, wt_u8(val>>16), wt_u16(val));
	}
	else
	{
		AddSingle(op_i, wt_u8(val));
	}
}

void InstrumentEngine::MakeENTER(wtast::FunctionSignature* sig)
{
	mInstruments.push_back(Instrument(Instrument::IK_ENTER, sig->GetCanonicalName(false, true)));
}

void InstrumentEngine::MakeEXEC(wtast::FunctionSignature* sig)
{
	map_uniqueid_ins_t::iterator it = mEXEC_Table.find(sig->GetUniqueId());

	if(mEXEC_Table.end()==it)
	{
		std::ostringstream ss;
		ss << "Unhandled builtin function: " <<
			sig->GetCanonicalName(true, true);
		mProblemHandler->InstrumentError(ss.str());
		return;
	}

	mInstruments.insert(mInstruments.end(), it->second.begin(), it->second.end());
}

CodeSection* InstrumentEngine::Compile(const std::string & cs_name, wt_u32 cs_addr, DLabelSection* dlabs, bool all_labels)
{
	CodeSection* const cs = new CodeSection(
		mProblemHandler->GetProblemHandler(), cs_name, cs_addr);

	if(!cs)
	{
		mProblemHandler->InstrumentError("Not enough memory for CodeSection");
		return 0;
	}

	wt_u32 base_pc = cs_addr / sizeof(wt_u16);

	map_pc_resolve_t resolve_enter;
	map_pc_resolve_t resolve_mkfp;
	map_pc_resolve_t resolve_jmpi;
	map_pc_resolve_t resolve_jmpl;
	map_pc_resolve_t resolve_jti;
	map_pc_resolve_t resolve_jtl;
	map_pc_resolve_t resolve_jfi;
	map_pc_resolve_t resolve_jfl;
	map_pc_resolve_t resolve_exhan;

	for(set_uniqueid_t::iterator it=mInstantiateBuiltins.begin(),end=mInstantiateBuiltins.end();it!=end;++it)
	{
		map_uniqueid_ins_t::iterator builtin_it = mEXEC_Table.find(*it);
		map_uniqueid_str_t::iterator name_it = mBuiltinNames.find(*it);

		if(mEXEC_Table.end()==builtin_it || mBuiltinNames.end()==name_it)
		{
			mProblemHandler->InstrumentError("Invalid entry in the builtin instantiation set.");
			continue;
		}

		AddGlobal(name_it->second);
		mInstruments.insert(mInstruments.end(), builtin_it->second.begin(), builtin_it->second.end());
		AddSingle(WT_TOMB_OP_LEAVE);
	}

	for(instruments_t::iterator it=mInstruments.begin(),end=mInstruments.end();it!=end;++it)
	{
		Instrument & ins = *it;

		switch(ins.kind)
		{
		case Instrument::IK_SINGLE:
			cs->AddInstruction(WT_TOMB_BUILDINSTR(ins.opcode, ins.imm8));
			break;

		case Instrument::IK_DOUBLE:
			cs->AddInstruction(WT_TOMB_BUILDINSTR(ins.opcode, ins.imm8));
			cs->AddInstruction(ins.imm16);
			break;

		case Instrument::IK_GLOBAL:
			cs->AddFunction(ins.data, base_pc + cs->GetNumInstructions());
		case Instrument::IK_LABEL:
			mMapLabelToPC[ins.data] = base_pc + cs->GetNumInstructions();
			if(dlabs && (all_labels || Instrument::IK_GLOBAL==ins.kind))
			{
				dlabs->AddLabel(cs_addr + cs->GetNumInstructions()*sizeof(wt_u16), ins.data);
			}
			break;

		case Instrument::IK_JMPI:
			resolve_jmpi[cs->GetNumInstructions()] = ins.data;
			cs->AddInstruction(0);
			break;

		case Instrument::IK_JMPL:
			resolve_jmpl[cs->GetNumInstructions()] = ins.data;
			cs->AddInstruction(0);
			cs->AddInstruction(0);
			break;

		case Instrument::IK_JTI:
			resolve_jti[cs->GetNumInstructions()] = ins.data;
			cs->AddInstruction(0);
			break;

		case Instrument::IK_JTL:
			resolve_jtl[cs->GetNumInstructions()] = ins.data;
			cs->AddInstruction(0);
			cs->AddInstruction(0);
			break;

		case Instrument::IK_JFI:
			resolve_jfi[cs->GetNumInstructions()] = ins.data;
			cs->AddInstruction(0);
			break;

		case Instrument::IK_JFL:
			resolve_jfl[cs->GetNumInstructions()] = ins.data;
			cs->AddInstruction(0);
			cs->AddInstruction(0);
			break;

		case Instrument::IK_ENTER:
			resolve_enter[cs->GetNumInstructions()] = ins.data;
			cs->AddInstruction(0);
			cs->AddInstruction(0);
			break;

		case Instrument::IK_MKFP:
			resolve_mkfp[cs->GetNumInstructions()] = ins.data;
			cs->AddInstruction(0);
			cs->AddInstruction(0);
			break;

		case Instrument::IK_EXHAN:
			resolve_exhan[cs->GetNumInstructions()] = ins.data;
			cs->AddInstruction(0);
			cs->AddInstruction(0);
			break;
		}
	}

	Resolve(cs, resolve_enter, WT_TOMB_OP_ENTER_L, false, false);
	Resolve(cs, resolve_mkfp,  WT_TOMB_OP_MKFP_L, false, false);
	Resolve(cs, resolve_exhan, WT_TOMB_OP_EXHAN_L, true, false);
	Resolve(cs, resolve_jmpi, WT_TOMB_OP_JMP_I, true, true);
	Resolve(cs, resolve_jmpl, WT_TOMB_OP_JMP_L, true, false);
	Resolve(cs, resolve_jti, WT_TOMB_OP_JT_I, true, true);
	Resolve(cs, resolve_jtl, WT_TOMB_OP_JT_L, true, false);
	Resolve(cs, resolve_jfi, WT_TOMB_OP_JF_I, true, true);
	Resolve(cs, resolve_jfl, WT_TOMB_OP_JF_L, true, false);

	return cs;
}

void InstrumentEngine::Resolve(CodeSection* cs, map_pc_resolve_t & resolve, wt_u8 opcode, bool relative, bool single)
{
	while(!resolve.empty())
	{
		std::string label = resolve.begin()->second;
		wt_u32 base = resolve.begin()->first;
		wt_u32 resd = mMapLabelToPC[label];
		resolve.erase(resolve.begin());

		if(0==resd)
		{
			mProblemHandler->InstrumentError(std::string("Failed to resolve \"") + label + "\"");
			continue;
		}

		if(relative)
		{
			resd = resd - base - cs->GetAddr() / sizeof(wt_u16);

			if((0!=(0xff000000ul & resd) && 0xff000000ul!=(0xff000000ul & resd)) || (single && (0!=(0xffffff00ul & resd) && 0xffffff00ul!=(0xffffff00ul & resd))))
			{
				mProblemHandler->InstrumentError(std::string("Internal error: resolved address of \"") + label + "\" is too large");
				return;
			}
		}
		else
		if((0xff000000ul & resd) || (single && (0xffffff00ul & resd)))
		{
			mProblemHandler->InstrumentError(std::string("Internal error: resolved address of \"") + label + "\" is too large");
			return;
		}

		if(single)
		{
			cs->mInstructions[base] = WT_TOMB_BUILDINSTR(opcode, wt_u8(resd));
		}
		else
		{
			cs->mInstructions[base+0] = WT_TOMB_BUILDINSTR(opcode, resd>>16);
			cs->mInstructions[base+1] = wt_u16(resd);
		}
	}
}

} // codegen
} // wtc
