//================================================================================================
// $Id: codegen_stdlib.cpp 7190 2024-01-08 18:40:25Z aron $
//
// Project:
//		Welltype Compiler
// Contents:
//		Code generator
// Authors:
//		Aron Barath, 2015-2023
//================================================================================================
// Copyright (c) Aron Barath
//================================================================================================

#include "codegen.h"
#include "tinyframe.h"
#include "welltype/wtruntime.h"
#include "wtruntime/types.h"
#include "welltype/stdlib.h"
#include "wtparser/ast/ast_type.h"
#include "wtparser/ast/ast_seqtype.h"
#include "wtparser/ast/ast_functionsignature.h"
#include "wtparser/ast/ast_functionargument.h"

namespace wtc
{
namespace codegen
{

void CodeGen::GenStdlib_strtok(wtast::FunctionSignature* sig)
{
	if(AddAssert(mProblemHandler,
		 2==sig->GetArgc() && 1==sig->GetRetc()
		&& WTB3T_STRING==sig->GetArg(0)->GetType()->GetTypeId()
		&& WTB3T_STRING==sig->GetArg(1)->GetType()->GetTypeId()
		&& sig->GetRet(0)->IsSequence() &&
		WTB3T_STRING==static_cast<wtast::SeqType*>(
			sig->GetRet(0))->GetElementType()->GetTypeId()))
	{
		return;
	}

	mEngine.AddGlobal(sig->GetCanonicalName(false, true));

	wtast::Type* const result_type = sig->GetRet(0);
	const std::string  lab_start   = mEngine.GenLabel();
	const std::string  lab_beg_cpy = mEngine.GenLabel();
	const std::string  lab_aft_cpy = mEngine.GenLabel();
	const std::string  lab_end     = mEngine.GenLabel();
	TinyFrame          tf;

	const wt_u32 o_string = tf.Add(sizeof(wt_vmptr));
	const wt_u32 o_sep    = tf.Add(sizeof(wt_vmptr));
	const wt_u32 o_result = tf.Add(sizeof(wt_vmptr));
	const wt_u32 o_offs   = tf.Add(sizeof(wt_u32));
	const wt_u32 o_len    = tf.Add(sizeof(wt_u32));
	const wt_u32 o_tmp    = tf.Add(sizeof(wt_u32));
	const wt_u32 o_oldseq = tf.Add(sizeof(wt_u32));
	const wt_u32 o_index  = tf.Add(sizeof(wt_u32));

	// allocate DVT and CSOS
	mEngine.MakeDVTA(tf.Size());
	mEngine.AddSingle(WT_TOMB_OP_OSA_I, 1);

	// store params
	mEngine.MakePVT(o_sep);
	mEngine.MakeST(sizeof(wt_vmptr));
	mEngine.MakePVT(o_string);
	mEngine.MakeST(sizeof(wt_vmptr));

	// create result seq
	mEngine.AddSingle(WT_TOMB_OP_PWZ_I, 0);
	mEngine.AddSingle(WT_TOMB_OP_SEQN_I, sizeof(wt_vmptr));
	mEngine.MakePVT(o_result);
	mEngine.MakeST(sizeof(wt_vmptr));
	mEngine.MakePVT(o_result);
	mEngine.AddMkfp(GetDestructorName(result_type));
	mEngine.AddSingle(WT_TOMB_OP_OSPU);

	// init 'offset'
	mEngine.MakePVT(o_string);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.MakePVT(o_sep);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.MakeEXEC(WTMP_STRSPN);
	mEngine.MakePVT(o_offs);
	mEngine.MakeST(sizeof(wt_u32));

	// calculate length of the input string
	mEngine.MakePVT(o_string);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_STRL);
	mEngine.MakePVT(o_len);
	mEngine.MakeST(sizeof(wt_u32));

	// loop start
	mEngine.AddLabel(lab_start);
	// while(offs<len)
	mEngine.MakePVT(o_offs);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.MakePVT(o_len);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.AddSingle(WT_TOMB_OP_CMP);
	mEngine.AddSingle(WT_TOMB_OP_QLT);
	mEngine.AddJFI(lab_end);

	// find the end of the current token
	mEngine.MakePVT(o_string);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.MakePVT(o_offs);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.MakePVT(o_sep);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.MakeEXEC(WTMP_STRCSPN2);
	mEngine.MakePVT(o_tmp);
	mEngine.MakeST(sizeof(wt_u32)); // tmp: end of the current token (offset from string start)

	// push string for substr
	mEngine.MakePVT(o_string);
	mEngine.MakeLD(sizeof(wt_vmptr));
	// push 'offset' for substr
	mEngine.MakePVT(o_offs);
	mEngine.MakeLD(sizeof(wt_u32));
	// calculate 'count' for substr
	mEngine.MakePVT(o_tmp);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.MakePVT(o_offs);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.AddSingle(WT_TOMB_OP_SUB_S);
	// call 'substr'
	mEngine.MakeEXEC(WTMP_STRSUB);

	// create a longer sequence
	mEngine.MakePVT(o_result);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.MakeDUP(sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_SEQL);
	mEngine.AddSingle(WT_TOMB_OP_ADD_I, 1);
	mEngine.AddSingle(WT_TOMB_OP_SEQN_I, sizeof(wt_vmptr));
	mEngine.MakePVT(o_result);
	mEngine.MakeST(sizeof(wt_vmptr));
	mEngine.MakePVT(o_oldseq);
	mEngine.MakeST(sizeof(wt_vmptr));
	// copy loop
	mEngine.AddSingle(WT_TOMB_OP_PWZ_I, 0);
	mEngine.MakePVT(o_index);
	mEngine.MakeST(sizeof(wt_u32));
	mEngine.MakePVT(o_oldseq);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_SEQL);
	// loop starts here
	mEngine.AddLabel(lab_beg_cpy);
	mEngine.MakeDUP(sizeof(wt_u32));
	mEngine.MakePVT(o_index);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.AddSingle(WT_TOMB_OP_CMP);
	mEngine.AddSingle(WT_TOMB_OP_QGT);
	mEngine.AddJFI(lab_aft_cpy);
	// load 'i'th string
	mEngine.MakePVT(o_oldseq);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.MakePVT(o_index);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.AddSingle(WT_TOMB_OP_SEQIR_S);
	mEngine.MakeLD(sizeof(wt_vmptr));
	// store 'i'th string
	mEngine.MakePVT(o_result);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.MakePVT(o_index);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.AddSingle(WT_TOMB_OP_SEQIR_S);
	mEngine.MakeST(sizeof(wt_vmptr));
	// increment index
	mEngine.MakePVT(o_index);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.AddSingle(WT_TOMB_OP_ADD_I, 1);
	mEngine.MakePVT(o_index);
	mEngine.MakeST(sizeof(wt_u32));
	// go back to the copy loop start
	mEngine.AddJmpI(lab_beg_cpy);
	// end of copy loop
	mEngine.AddLabel(lab_aft_cpy);
	mEngine.AddSingle(WT_TOMB_OP_POP_I, sizeof(wt_u32));

	// store next token
	mEngine.MakePVT(o_result);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.MakePVT(o_index);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.AddSingle(WT_TOMB_OP_SEQIR_S);
	mEngine.MakeST(sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_TDR);

	// delete the old sequence
	mEngine.MakePVT(o_oldseq);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_SEQD);

	// find the end of the current token
	mEngine.MakePVT(o_string);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.MakePVT(o_tmp);
	mEngine.MakeLD(sizeof(wt_u32));
	mEngine.MakePVT(o_sep);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.MakeEXEC(WTMP_STRSPN2);
	mEngine.MakePVT(o_offs);
	mEngine.MakeST(sizeof(wt_u32)); // tmp: end of the current token (offset from string start)

	// jump
	mEngine.AddJmpI(lab_start);

	// end
	mEngine.AddLabel(lab_end);
	// finalize
	mEngine.AddSingle(WT_TOMB_OP_OSDR);
	mEngine.MakePVT(o_result);
	mEngine.MakeLD(sizeof(wt_vmptr));
	mEngine.AddSingle(WT_TOMB_OP_DUP_I, sizeof(wt_vmptr));
	mEngine.AddMkfp(GetDestructorName(result_type));
	mEngine.AddSingle(WT_TOMB_OP_TPU);
	mEngine.AddSingle(WT_TOMB_OP_OSF);
	mEngine.AddSingle(WT_TOMB_OP_DVTF);
	mEngine.AddSingle(WT_TOMB_OP_LEAVE);
}

} // codegen
} // wtc
