#include "config_private.h"
#include "loader3.h"
#include "lreader.h"
#include "program.h"
#include "vm.h"
#include "ctx.h"
#include "hash.h"
#include "version.h"
#include "sym.h"
#include "sym3.h"
#include "symtab.h"
#include "symtab3.h"
#include <string.h>

#define CHECK(_cond_)           CHECK_ERR(_cond_, WTE_BINERR)

#define CHECK_ERR(_cond_,_err_) \
	do \
	{ \
		if(!(_cond_)) \
		{ \
			wtr_loader3_check_failed_impl(vm, __LINE__); \
			err = (_err_); \
			goto error_1; \
		} \
	} \
	while(0)

static void wtr_loader3_check_failed_impl(struct wtr_vm* const vm,
	unsigned int line)
{
	if(NULL!=vm->loaderops)
	{
		void (*const check_failed)(void* ctx, unsigned int line) =
			vm->loaderops->check_failed;

		if(NULL!=check_failed)
		{
			check_failed(vm->loaderops_ctx, line);
		}
	}
}

static wt_error wtr_loader3_bos2(struct wtr_vm* const vm, wt_vmptr data,
	wt_uint32_t cnt)
{
	while(0<(cnt--))
	{
		wt_uint16_t tmp;

		if(0!=wtr_vma_read(vm, data, &tmp, sizeof(tmp)))
		{
			return WTE_FAULT;
		}

		/* this function is called only when WTB and host byteorder is */
		/* different, thus we MUST swap */
		tmp = wtr_swap16(tmp);

		if(0!=wtr_vma_write(vm, data, &tmp, sizeof(tmp)))
		{
			return WTE_FAULT;
		}

		data += sizeof(tmp);
	}

	return WTE_OK;
}

static wt_error wtr_loader3_bos4(struct wtr_vm* const vm, wt_vmptr data,
	wt_uint32_t cnt)
{
	while(0<(cnt--))
	{
		wt_uint32_t tmp;

		if(0!=wtr_vma_read(vm, data, &tmp, sizeof(tmp)))
		{
			return WTE_FAULT;
		}

		/* this function is called only when WTB and host byteorder is */
		/* different, thus we MUST swap */
		tmp = wtr_swap32(tmp);

		if(0!=wtr_vma_write(vm, data, &tmp, sizeof(tmp)))
		{
			return WTE_FAULT;
		}

		data += sizeof(tmp);
	}

	return WTE_OK;
}

static wt_error wtr_loader3_bos8(struct wtr_vm* const vm, wt_vmptr data,
	wt_uint32_t cnt)
{
	while(0<(cnt--))
	{
		wt_uint64_t tmp;

		if(0!=wtr_vma_read(vm, data, &tmp, sizeof(tmp)))
		{
			return WTE_FAULT;
		}

		/* this function is called only when WTB and host byteorder is */
		/* different, thus we MUST swap */
		tmp = wtr_swap64(tmp);

		if(0!=wtr_vma_write(vm, data, &tmp, sizeof(tmp)))
		{
			return WTE_FAULT;
		}

		data += sizeof(tmp);
	}

	return WTE_OK;
}

static int wtr_loader3_add_metadata(struct wtr_vm* const vm,
	struct wtr_program* const prog, struct wtr_metadata* const metadata)
{
	if(0!=wtr_vector_push_back(&prog->metadata_vec, vm, metadata))
	{
		wtr_vm_sys_free(vm, metadata);
		return 1;
	}

	return 0;
}

static wt_error wtr_loader3_load_code(struct wtr_vm* const vm,
	struct wtr_lreader* const lr, wt_vmptr dst, wt_uint32_t offs,
	wt_uint32_t size)
{
	const wt_uint32_t end = offs + size;

	while(offs<end)
	{
		wt_uint16_t instr = lr->read16(lr, offs);

		if(0!=wtr_vma_write(vm, dst, &instr, sizeof(wt_uint16_t)))
		{
			return WTE_FAULT;
		}

		dst  += sizeof(wt_uint16_t);
		offs += sizeof(wt_uint16_t);
	}

	return WTE_OK;
}

static wt_error wtr_loader3_load_shdr(struct wtr_lreader* const lr,
	wt_uint32_t offs, struct wtb3_shdr* const shdr)
{
	shdr->sh_type = lr->read8(lr, offs+offsetof(struct wtb3_shdr, sh_type));
	shdr->sh_flag = lr->read8(lr, offs+offsetof(struct wtb3_shdr, sh_flag));
	shdr->sh_link = lr->read16(lr, offs+offsetof(struct wtb3_shdr, sh_link));
	shdr->sh_offs = lr->read32(lr, offs+offsetof(struct wtb3_shdr, sh_offs));
	shdr->sh_size = lr->read32(lr, offs+offsetof(struct wtb3_shdr, sh_size));
	shdr->sh_addr = lr->read32(lr, offs+offsetof(struct wtb3_shdr, sh_addr));

	return WTE_OK;
}

#define RELOCS_PER_NODE 4

struct wtr_reloc3
{
	enum wtr_symtype symtype;
	wt_uint32_t      sym_idx;
	wt_uint32_t      offs;
};

struct wtr_relocnode3
{
	struct wtr_relocnode3* prev;
	size_t                 count;
	struct wtr_reloc3      relocs[RELOCS_PER_NODE];
};

static wt_error wtr_loader3_add_reloc(struct wtr_vm* const vm,
	struct wtr_relocnode3** const p_head, enum wtr_symtype symtype,
	size_t idx, wt_uint32_t offs)
{
	struct wtr_relocnode3* node = *p_head;
	struct wtr_reloc3*     rel3;

	if(NULL==node || RELOCS_PER_NODE==node->count)
	{
		struct wtr_relocnode3* const new_node = (struct wtr_relocnode3*)
			wtr_vm_sys_malloc(vm, sizeof(struct wtr_relocnode3));

		if(NULL==new_node)
		{
			return WTE_NOMEM;
		}

		memset(new_node, 0, sizeof(struct wtr_relocnode3));

		new_node->prev = node;
		*p_head = new_node;
		node = new_node;
	}

	rel3 = &node->relocs[node->count++];

	rel3->symtype = symtype;
	rel3->sym_idx = idx;
	rel3->offs    = offs;

	return WTE_OK;
}

static wt_error wtr_loader3_apply_relocation(struct wtr_vm* const vm,
	struct wtr_program* const prog, struct wtr_symtab3* const symtab,
	struct wtr_reloc3* const rel3)
{
	switch(rel3->symtype)
	{
	case WTR_SYMTYPE_EXC:
		{
			struct wtr_sym_exc* const exc = (struct wtr_sym_exc*)
				wtr_symtab3_get_symbol(symtab, rel3->sym_idx);

			if(NULL!=exc && WTR_SYMTYPE_EXC==exc->sym.type)
			{
				wt_uint32_t value = exc->exc_id;

				if(0!=wtr_vma_write(vm, prog->program_data+rel3->offs, &value,
					sizeof(value)))
				{
					return WTE_FAULT;
				}

				return WTE_OK;
			}
		}
		break;

	case WTR_SYMTYPE_WTFUNC:
		{
			struct wtr_sym* const sym = wtr_symtab3_get_symbol(symtab,
				rel3->sym_idx);

			if(NULL!=sym)
			{
				struct wtr_sym* const def = wtr_gidtree_get(&vm->gidtree,
					((struct wtr_sym_func*)sym)->func_id);

				if(NULL!=def)
				{
					struct wt_pfn value;

					switch(def->type)
					{
					case WTR_SYMTYPE_WTFUNC:
						{
							struct wtr_sym_wtfunc* const func =
								(struct wtr_sym_wtfunc*)def;

							value.ptr = def->symtab->prog;
							value.ext = func->entry;
						}
						break;

					case WTR_SYMTYPE_CFUNC:
						{
							struct wtr_sym_cfunc* const func =
								(struct wtr_sym_cfunc*)def;

							value.ptr = func;
							value.ext = vm->external_tag;
						}
						break;

					default:
						return WTE_BADREL;
					}

					if(0!=wtr_vma_write(vm, prog->program_data+rel3->offs,
						&value, sizeof(value)))
					{
						return WTE_FAULT;
					}

					return WTE_OK;
				}
			}
		}
		break;

	default:
		break;
	}

	return WTE_BADREL;
}

wt_error wtr_loader3(struct wtr_vm* const vm,
	struct wtr_program** const p_program, struct wtr_context* const ctx,
	struct wtr_lreader* const lr, wt_uint32_t flags,
	const struct wtb3_hdr* const hdr)
{
	struct wtr_program* prog;
	struct wtr_symtab3* symtab = NULL;

	struct wtr_relocnode3* relocs = NULL;

	wt_error    err     = WTE_OK;
	wt_uint32_t b_size  = 0;
	wt_uint16_t b_shnum = 0;
	wt_uint16_t b_thnum = 0;
	wt_uint32_t b_shoff = 0;
	wt_uint32_t b_thoff = 0;
	wt_uint32_t b_ctor  = 0;
	wt_uint32_t b_dtor  = 0;
	wt_bo       bo;

	if(sizeof(struct wtb3_hdr)>lr->size)
	{
		return WTE_BINERR;
	}

	if(WTB3MAG0!=hdr->b_ident[WTB3I_MAGIC0] ||
		WTB3MAG1!=hdr->b_ident[WTB3I_MAGIC1] ||
		WTB3MAG2!=hdr->b_ident[WTB3I_MAGIC2] ||
		WTB3MAG3!=hdr->b_ident[WTB3I_MAGIC3])
	{
		return WTE_BADMAG;
	}

	if(WTB3V_CURRENT!=hdr->b_ident[WTB3I_VERSION])
	{
		return WTE_BADVER;
	}

	bo = hdr->b_ident[WTB3I_BYTEORDER];
	if(WTBO_LITTLE!=bo && WTBO_BIG!=bo)
	{
		return WTE_BINERR;
	}

	{
		/* remnants of the 4-component version number, there is a trick */
		/* with WTB3I_COMPVER3: it must be zero since v0.9.5 -- this will */
		/* be fixed in WTB4 */

		wt_uint32_t h_ver = WT_MAKEVERSION(
			hdr->b_ident[WTB3I_COMPVER0],
			hdr->b_ident[WTB3I_COMPVER1],
			hdr->b_ident[WTB3I_COMPVER2]);

		if(WT_MAKEVERSION(0, 9, 5)<=h_ver && 0!=hdr->b_ident[WTB3I_COMPVER3])
		{
			return WTE_BINERR;
		}

		h_ver |= hdr->b_ident[WTB3I_COMPVER3];

		if(WTR_BINCOMPSINCEVER>h_ver || WTR_VERSION<h_ver)
		{
			return WTE_ABIVER;
		}
	}

	b_size  = lr->read32(lr, offsetof(struct wtb3_hdr, b_size));
	b_shnum = lr->read16(lr, offsetof(struct wtb3_hdr, b_shnum));
	b_thnum = lr->read16(lr, offsetof(struct wtb3_hdr, b_thnum));
	b_shoff = lr->read32(lr, offsetof(struct wtb3_hdr, b_shoff));
	b_thoff = lr->read32(lr, offsetof(struct wtb3_hdr, b_thoff));
	b_ctor  = lr->read32(lr, offsetof(struct wtb3_hdr, b_ctor));
	b_dtor  = lr->read32(lr, offsetof(struct wtb3_hdr, b_dtor));

	if(0==b_size || 0==b_shnum || 0==b_shoff || 0==b_thoff)
	{
		return WTE_BINERR;
	}

	if(1!=b_thnum)
	{
		return WTE_BINERR;
	}

	if(NULL==(prog=(struct wtr_program*)wtr_vm_sys_malloc(vm,
		sizeof(struct wtr_program))))
	{
		return WTE_NOMEM;
	}

	memset(prog, 0, sizeof(*prog));

	if(0==(prog->program_data=wtr_vm_vma_malloc(vm, b_size)))
	{
		wtr_vm_sys_free(vm, prog);
		wtr_loader3_check_failed_impl(vm, __LINE__);
		return WTE_NOMEM;
	}

	if(0!=wtr_vma_memset(vm, prog->program_data, 0, b_size))
	{
		err = WTE_FAULT;
		goto error_0;
	}

	if(0!=wtr_vector_create(&prog->referenced_symtabs, vm, 4))
	{
		err = WTE_NOMEM;
		goto error_0;
	}

	if(0!=wtr_vector_create(&prog->metadata_vec, vm, 4))
	{
		wtr_vector_delete(&prog->referenced_symtabs, vm);
		err = WTE_NOMEM;
		goto error_0;
	}

	prog->prog_id      = vm->next_prog_id++;
	prog->refcount     = 0;
	prog->vm           = vm;
	prog->program_size = b_size;
	prog->pc_ctor      = b_ctor;
	prog->pc_dtor      = b_dtor;
	prog->pc_main      = WTR_NULLPC;
	prog->pc_base      = prog->program_data / 2;
	prog->base_address = prog->program_data;

	/* validate section headers, fetch typevec size and perform integrity */
	/* check */
	{
		wt_uint32_t      n_integ = 0;
		wt_uint32_t      typecnt = 0;
		wt_uint32_t      shoff   = b_shoff;
		wt_uint16_t      i;
		struct wtb3_shdr shdr;

		if(WTE_OK!=(err=wtr_loader3_load_shdr(lr, shoff, &shdr)))
		{
			wtr_loader3_check_failed_impl(vm, __LINE__);
			goto error_1;
		}

		shoff += sizeof(shdr);

		CHECK(WTB3ST_UNUSED==shdr.sh_type);

		for(i=1;i<b_shnum;shoff+=sizeof(shdr),++i)
		{
			if(WTE_OK!=(err=wtr_loader3_load_shdr(lr, shoff, &shdr)))
			{
				wtr_loader3_check_failed_impl(vm, __LINE__);
				goto error_1;
			}

			CHECK(shdr.sh_offs<lr->size &&
				shdr.sh_offs+shdr.sh_size<=lr->size);
			CHECK(0==(WTB3SF_LOAD & shdr.sh_flag) ||
				(shdr.sh_addr<b_size && shdr.sh_addr+shdr.sh_size<=b_size));

			switch(shdr.sh_type)
			{
			case WTB3ST_IMPORT:
			case WTB3ST_EXPORT:
				CHECK((sizeof(struct wtb3_symhdr)+sizeof(struct wtb3_sym))
					< shdr.sh_size);
				break;

			case WTB3ST_CODE:
				CHECK(0==shdr.sh_link);
				CHECK(0==(shdr.sh_size % sizeof(wt_uint16_t)));
				CHECK(0==(shdr.sh_offs % sizeof(wt_uint16_t)));
				CHECK(WTB3SF_LOAD & shdr.sh_flag);
				break;

			case WTB3ST_BOS:
				{
					struct wtb3_shdr link_shdr;

					CHECK(shdr.sh_link<b_shnum);

					if(WTE_OK!=(err=wtr_loader3_load_shdr(lr,
						b_shoff+shdr.sh_link*sizeof(shdr), &link_shdr)))
					{
						wtr_loader3_check_failed_impl(vm, __LINE__);
						goto error_1;
					}

					CHECK(link_shdr.sh_link==i);
					CHECK(link_shdr.sh_type==WTB3ST_DATA);
					CHECK(link_shdr.sh_addr==shdr.sh_addr);
				}
				break;

			case WTB3ST_METADATA:
				{
					CHECK(sizeof(struct wtb3_metahdr)<shdr.sh_size);
					CHECK(0==shdr.sh_flag);

					/* last octet must be 0 */
					CHECK(0==lr->read8(lr, shdr.sh_offs+shdr.sh_size-1));
				}
				break;

			case WTB3ST_DEBUGLABEL:
				{
					/* just a sanity check */

					CHECK(sizeof(struct wtb3_dlabelhdr)<shdr.sh_size);
					CHECK(0==shdr.sh_flag);

					/* last octet must be 0 */
					CHECK(0==lr->read8(lr, shdr.sh_offs+shdr.sh_size-1));
				}
				break;

			case WTB3ST_DATA:
				if(WTB3SF_TYPEVEC & shdr.sh_flag)
				{
					CHECK(0==typecnt);
					typecnt = shdr.sh_size / sizeof(wt_handle);
				}

				if(0!=shdr.sh_link)
				{
					struct wtb3_shdr link_shdr;

					CHECK(shdr.sh_link<b_shnum);

					if(WTE_OK!=(err=wtr_loader3_load_shdr(lr,
						b_shoff+shdr.sh_link*sizeof(shdr), &link_shdr)))
					{
						wtr_loader3_check_failed_impl(vm, __LINE__);
						goto error_1;
					}

					CHECK(link_shdr.sh_link==i);
					CHECK(link_shdr.sh_type==WTB3ST_BOS);
				}
				break;

			case WTB3ST_INTEGRITY:
				{
					struct wtb3_integhdr ihdr;

					const struct wtr_hashalg* alg;

					ihdr.algorithm = lr->read16(lr, shdr.sh_offs+
						offsetof(struct wtb3_integhdr, algorithm));
					ihdr.reserved = lr->read16(lr, shdr.sh_offs+
						offsetof(struct wtb3_integhdr, reserved));

					CHECK(shdr.sh_size>sizeof(ihdr));
					CHECK(0==shdr.sh_flag);
					CHECK(0==ihdr.reserved);

					/* it is allowed to skip unknown/unsupported hash */
					/* algorithms -- the only important thing is to check */
					/* what we can, and later evaluate the REQINTEGRITY */
					/* requirements */

					if(NULL!=(alg=wtr_hash_find_by_id(vm->hashalgs,
						ihdr.algorithm)))
					{
						/* no dangling data is allowed in INTEGRITY section */
						CHECK(alg->digest_len+sizeof(ihdr)==shdr.sh_size);

						if(WTE_OK!=(err=wtr_hash_check_lr(lr, alg,
							shdr.sh_offs+sizeof(ihdr))))
						{
							wtr_loader3_check_failed_impl(vm, __LINE__);
							goto error_1;
						}

						++n_integ;
					}
				}
				break;

			default:
				wtr_loader3_check_failed_impl(vm, __LINE__);
				err = WTE_BADVER;
				goto error_1;
			}
		}

		/* check REQINTEGRITY */

		if((WTRVMF_REQINTEGRITY & vm->accepted_flags) ||
			(WTRLF_REQINTEGRITY & flags))
		{
			/* there is an integrity check requirement, thus n_integ must */
			/* not be zero */

			if(0==n_integ)
			{
				wtr_loader3_check_failed_impl(vm, __LINE__);
				err = WTE_INTEG;
				goto error_1;
			}
		}

		/* create symbol table */

		if(NULL==(symtab=wtr_symtab3_create(vm, prog, typecnt)))
		{
			wtr_loader3_check_failed_impl(vm, __LINE__);
			err = WTE_NOMEM;
			goto error_1;
		}

		prog->symtab = (struct wtr_symtab*)symtab;
	}

	/* process program data sections (CODE, DATA, IMPORT, EXPORT) */
	{
		wt_uint32_t      shoff = b_shoff + sizeof(struct wtb3_shdr);
		wt_uint16_t      i;
		struct wtb3_shdr shdr;

		for(i=1;i<b_shnum;shoff+=sizeof(shdr),++i)
		{
			if(WTE_OK!=(err=wtr_loader3_load_shdr(lr, shoff, &shdr)))
			{
				goto error_1;
			}

			switch(shdr.sh_type)
			{
			case WTB3ST_EXPORT:
			case WTB3ST_IMPORT:
				{
					wt_uint32_t offs = shdr.sh_offs;

					wt_uint32_t symcnt = lr->read32(lr,
						offs+offsetof(struct wtb3_symhdr, symcount));

					CHECK(0==lr->read32(lr, offs+offsetof(struct wtb3_symhdr,
						reserved)));
					CHECK(0!=symcnt);

					offs += sizeof(struct wtb3_symhdr);

					for(;symcnt>0;offs+=sizeof(struct wtb3_sym),--symcnt)
					{
						struct wtb3_sym sym;
						struct wtr_sym* rt_sym;
						enum wtr_symtype symtype;
						enum wtr_symbind symbind;
						size_t sym_idx;

						sym.sym_type = lr->read32(lr,
							offs+offsetof(struct wtb3_sym, sym_type));
						sym.sym_name = lr->read32(lr,
							offs+offsetof(struct wtb3_sym, sym_name));
						sym.sym_offs = lr->read32(lr,
							offs+offsetof(struct wtb3_sym, sym_offs));
						sym.sym_info = lr->read32(lr,
							offs+offsetof(struct wtb3_sym, sym_info));

						CHECK(0!=sym.sym_name);

						if(WTB3ST_IMPORT==shdr.sh_type)
						{
							symbind = WTR_SYMBIND_IMPORT;
						}
						else
						{
							symbind = WTR_SYMBIND_EXPORT;
						}

						if(WTE_OK!=(err=wtr_sym3_parse(vm, lr,
							(struct wtr_symtab*)symtab, &rt_sym, shdr.sh_offs,
							&sym, symbind)))
						{
							wtr_loader3_check_failed_impl(vm, __LINE__);
							goto error_1;
						}

						symtype = rt_sym->type;

						switch(symtype)
						{
						case WTR_SYMTYPE_DATATYPE:
						case WTR_SYMTYPE_REC:
						case WTR_SYMTYPE_PRIVREC:
						case WTR_SYMTYPE_LIMREC:
						case WTR_SYMTYPE_FUNCTYPE:
						case WTR_SYMTYPE_ENUM:
							sym_idx = sym.sym_info;
							err = wtr_symtab3_add_type(symtab, vm, rt_sym,
								sym.sym_info);
							break;

						case WTR_SYMTYPE_WTFUNC:
							if(WTB3ST_EXPORT==shdr.sh_type)
							{
								((struct wtr_sym_wtfunc*)rt_sym)->entry =
									prog->pc_base + sym.sym_info;
							}
							/* fallthrough */
						default:
							err = wtr_symtab3_add_symbol(symtab, vm, rt_sym,
								&sym_idx);
							break;
						}

						if(WTE_OK!=err)
						{
							wtr_loader3_check_failed_impl(vm, __LINE__);
							wtr_sym_release(vm, rt_sym);
							goto error_1;
						}

						if(WTB3ST_IMPORT==shdr.sh_type &&
							(WTR_SYMTYPE_EXC==symtype ||
								WTR_SYMTYPE_WTFUNC==symtype))
						{
							if(WTE_OK!=(err=wtr_loader3_add_reloc(vm,
								&relocs, symtype, sym_idx, sym.sym_info)))
							{
								wtr_loader3_check_failed_impl(vm, __LINE__);
								goto error_1;
							}
						}
					}
				}
				break;

			case WTB3ST_DATA:
				{
					wt_vmptr data = prog->program_data + shdr.sh_addr;

					if(WTB3SF_CLEAR & shdr.sh_flag)
					{
						if(0!=wtr_vma_memset(vm, data, 0, shdr.sh_size))
						{
							err = WTE_FAULT;
							goto error_1;
						}
					}

					if(WTB3SF_LOAD & shdr.sh_flag)
					{
						if(shdr.sh_size!=lr->read_vma(lr, data, shdr.sh_offs,
							shdr.sh_size))
						{
							err = WTE_FAULT;
							goto error_1;
						}
					}
				}
				break;

			case WTB3ST_CODE:
				if(0!=(err=wtr_loader3_load_code(vm, lr,
					prog->program_data+shdr.sh_addr, shdr.sh_offs,
					shdr.sh_size)))
				{
					goto error_1;
				}
				break;
			}
		}
	}

	/* process auxiliary sections (BOS, METADATA) */
	{
		wt_uint32_t      shoff = b_shoff + sizeof(struct wtb3_shdr);
		wt_uint16_t      i;
		struct wtb3_shdr shdr;

		for(i=1;i<b_shnum;shoff+=sizeof(shdr),++i)
		{
			if(WTE_OK!=(err=wtr_loader3_load_shdr(lr, shoff, &shdr)))
			{
				goto error_1;
			}

			switch(shdr.sh_type)
			{
			case WTB3ST_METADATA:
				{
					wt_uint32_t i;
					wt_uint32_t offs = shdr.sh_offs;

					const wt_uint32_t cnt = lr->read32(lr,
						offs+offsetof(struct wtb3_metahdr, metacount));

					CHECK(0==lr->read32(lr,
						offs+offsetof(struct wtb3_metahdr, reserved)));
					CHECK(cnt<shdr.sh_size/sizeof(struct wtb3_meta));

					offs += sizeof(struct wtb3_metahdr);

					for(i=0;i<cnt;offs+=sizeof(struct wtb3_meta),++i)
					{
						const wt_uint32_t key_offs = lr->read32(lr,
							offs+offsetof(struct wtb3_meta, meta_key_offs));
						const wt_uint32_t val_offs = lr->read32(lr,
							offs+offsetof(struct wtb3_meta, meta_val_offs));

						wt_uint32_t key_size = wtr_lreader_strlen(lr,
							shdr.sh_offs+key_offs) + 1;
						wt_uint32_t val_size = wtr_lreader_strlen(lr,
							shdr.sh_offs+val_offs) + 1;

						struct wtr_metadata* md;
						char* str;

						CHECK(key_offs<shdr.sh_size);
						CHECK(val_offs<shdr.sh_size);

						md = (struct wtr_metadata*)wtr_vm_sys_malloc(vm,
							sizeof(*md)+key_size+val_size);

						CHECK_ERR(NULL!=md, WTE_NOMEM);

						str = (char*)(md + 1);
						md->key = str;
						CHECK(key_size==lr->read_sys(lr, str,
							shdr.sh_offs+key_offs, key_size));

						str += key_size;
						md->val = str;
						CHECK(val_size==lr->read_sys(lr, str,
							shdr.sh_offs+val_offs, val_size));

						CHECK_ERR(0==wtr_loader3_add_metadata(vm, prog, md),
							WTE_NOMEM);
					}
				}
				break;

			case WTB3ST_BOS:
				{
					wt_vmptr       data = prog->program_data + shdr.sh_addr;
					wt_vmptr       bos  = shdr.sh_offs;
					const wt_vmptr end  = bos + shdr.sh_size;

					while(bos<end)
					{
						const wt_uint8_t op = lr->read8(lr, bos++);

						wt_uint8_t  swap_sz;
						wt_uint8_t  cnt_sz;
						wt_uint64_t cnt;

						if(WTB3BOS_NOOP==op)
						{
							continue;
						}

						switch(op)
						{
						case WTB3BOS_NOSWAP_8:  swap_sz = 1; cnt_sz = 1; break;
						case WTB3BOS_NOSWAP_16: swap_sz = 1; cnt_sz = 2; break;
						case WTB3BOS_NOSWAP_32: swap_sz = 1; cnt_sz = 4; break;
						case WTB3BOS_NOSWAP_64: swap_sz = 1; cnt_sz = 8; break;
						case WTB3BOS_SWAP16_8:  swap_sz = 2; cnt_sz = 1; break;
						case WTB3BOS_SWAP16_16: swap_sz = 2; cnt_sz = 2; break;
						case WTB3BOS_SWAP16_32: swap_sz = 2; cnt_sz = 4; break;
						case WTB3BOS_SWAP16_64: swap_sz = 2; cnt_sz = 8; break;
						case WTB3BOS_SWAP32_8:  swap_sz = 4; cnt_sz = 1; break;
						case WTB3BOS_SWAP32_16: swap_sz = 4; cnt_sz = 2; break;
						case WTB3BOS_SWAP32_32: swap_sz = 4; cnt_sz = 4; break;
						case WTB3BOS_SWAP32_64: swap_sz = 4; cnt_sz = 8; break;
						case WTB3BOS_SWAP64_8:  swap_sz = 8; cnt_sz = 1; break;
						case WTB3BOS_SWAP64_16: swap_sz = 8; cnt_sz = 2; break;
						case WTB3BOS_SWAP64_32: swap_sz = 8; cnt_sz = 4; break;
						case WTB3BOS_SWAP64_64: swap_sz = 8; cnt_sz = 8; break;
						default:
							wtr_loader3_check_failed_impl(vm, __LINE__);
							err = WTE_BINERR;
							goto error_1;
						}

						switch(cnt_sz)
						{
						case 1: cnt = lr->read8(lr, bos); break;
						case 2: cnt = lr->read16(lr, bos); break;
						case 4: cnt = lr->read32(lr, bos); break;
						case 8: cnt = lr->read64(lr, bos); break;
						}

						bos += cnt_sz;

						CHECK(cnt*swap_sz<(wt_uint64_t)b_size);

						if(1<swap_sz && wtr_byteorder()!=bo)
						{
							switch(swap_sz)
							{
							case 2:
								err = wtr_loader3_bos2(vm, data, cnt);
								break;

							case 4:
								err = wtr_loader3_bos4(vm, data, cnt);
								break;

							case 8:
								err = wtr_loader3_bos8(vm, data, cnt);
								break;
							}

							if(WTE_OK!=err)
							{
								wtr_loader3_check_failed_impl(vm, __LINE__);
								goto error_1;
							}
						}

						data += cnt * swap_sz;
					}
				}
				break;
			}
		}
	}

	if(WTE_OK!=(err=wtr_symtab3_resolve_undef(symtab, vm)))
	{
		wtr_loader3_check_failed_impl(vm, __LINE__);
		goto error_1;
	}

	while(NULL!=relocs)
	{
		struct wtr_relocnode3* const prev = relocs->prev;

		size_t i;

		for(i=0;i<relocs->count;++i)
		{
			err = wtr_loader3_apply_relocation(vm, prog, symtab,
				&relocs->relocs[i]);

			if(WTE_OK!=err)
			{
				wtr_loader3_check_failed_impl(vm, __LINE__);
				goto error_1;
			}
		}

		wtr_vm_sys_free(vm, relocs);
		relocs = prev;
	}

	/* process our only thread header */
	{
		const wt_uint16_t th_type = lr->read16(lr,
			b_thoff + offsetof(struct wtb3_thdr, th_type));
		const wt_uint16_t th_flag = lr->read16(lr,
			b_thoff + offsetof(struct wtb3_thdr, th_flag));
		const wt_uint32_t th_cssize = lr->read32(lr,
			b_thoff + offsetof(struct wtb3_thdr, th_cssize));
		const wt_uint32_t th_dssize = lr->read32(lr,
			b_thoff + offsetof(struct wtb3_thdr, th_dssize));
		const wt_uint32_t th_pcbase = lr->read32(lr,
			b_thoff + offsetof(struct wtb3_thdr, th_pcbase));

		CHECK(0<th_cssize);
		CHECK(0<th_dssize);
		CHECK(WTB3TT_USER==th_type);
		CHECK(!(WTB3TF_EXECUTE & th_flag));

		prog->pc_main = th_pcbase;
	}

	/* free up some memory */

	wtr_vector_shrink_to_fit(&prog->referenced_symtabs, vm);
	wtr_vector_shrink_to_fit(&prog->metadata_vec, vm);

	/* call constructor if needed */

	if(WTR_NULLPC!=prog->pc_ctor)
	{
		wtr_ctx_invoke_ctor(ctx, prog);
	}

	/* done */

	wtr_list_insert_tail(&vm->list_programs, &prog->link_programs);
	wtr_list_insert_tail(&vm->list_symtabs, &prog->symtab->link_symtabs);

	*p_program = prog;
	++prog->refcount;

	return WTE_OK;

error_1:
	wtr_program_unref_symtabs(prog);
	wtr_program_free_metadata(prog);
	wtr_vector_delete(&prog->referenced_symtabs, vm);
	wtr_vector_delete(&prog->metadata_vec, vm);

	if(NULL!=prog->symtab)
	{
		prog->symtab->destroy(prog->symtab, vm);
	}

	while(NULL!=relocs)
	{
		struct wtr_relocnode3* const prev = relocs->prev;

		wtr_vm_sys_free(vm, relocs);
		relocs = prev;
	}

error_0:
	wtr_vm_vma_free(vm, prog->program_data);
	wtr_vm_sys_free(vm, prog);
	return err;
}
