#include "config_private.h"
#include "symtab_b.h"
#include "symtab.h"
#include "sym.h"
#include "vm.h"
#include "stdlib.h"
#include "vector.h"
#include <string.h>

struct wtr_symtab_b
{
	struct wtr_symtab symtab;
	struct wtr_vector usersyms;
};

#define WTR_DEF_EXCSYM(exc) \
static const struct wtr_sym_exc wtr_excsym_##exc = \
{ \
	{ \
		WTR_SYMTYPE_EXC, \
		WTR_SYMBIND_EXPORT, \
		NULL, \
		#exc \
	}, \
	exc \
};

#define WTR_EXCSYM_ELEM(exc) &wtr_excsym_##exc,

WTR_STDEXCEPTIONS(WTR_DEF_EXCSYM)

static const struct wtr_sym_exc* wtr_excsyms[] =
{
	WTR_STDEXCEPTIONS(WTR_EXCSYM_ELEM)
	NULL
};

static struct wtr_sym* wtr_symtab_b_lookup(struct wtr_symtab* const symtab0,
	struct wtr_vm* const vm, struct wtr_sym* const sym)
{
	struct wtr_symtab_b* const symtab =
		(struct wtr_symtab_b*)symtab0;

	if(WTR_SYMTYPE_EXC==sym->type)
	{
		const char* const sym_name = sym->name;

		const struct wtr_sym_exc** p_exc = wtr_excsyms;

		while(NULL!=(*p_exc))
		{
			const struct wtr_sym_exc* const exc = *p_exc++;

			if(0==strcmp(sym_name, exc->sym.name))
			{
				return (struct wtr_sym*)exc;
			}
		}
	}

	if(wtr_vector_size(&symtab->usersyms))
	{
		struct wtr_sym** syms = (struct wtr_sym**)
			wtr_vector_at(&symtab->usersyms, 0);

		struct wtr_sym** const end =
			syms + wtr_vector_size(&symtab->usersyms);

		while(syms<end)
		{
			struct wtr_sym* const my_sym = *syms++;

			if(0==wtr_sym_cmp(vm, my_sym, sym))
			{
				return my_sym;
			}
		}
	}

	return NULL;
}

/* NOTE: the builtin symtab is destroyed when the VM itself is destroyed, */
/* thus this can do a bit less (i.e. it won't remove symbols from GID), */
/* but releases system memory */
static void wtr_symtab_b_destroy(struct wtr_symtab* const symtab0,
	struct wtr_vm* const vm)
{
	struct wtr_symtab_b* const symtab = (struct wtr_symtab_b*)symtab0;

	wtr_symtab_symvec_destroy(vm, &symtab->usersyms);
	wtr_vm_sys_free(vm, symtab0);
}

struct wtr_symtab_b* wtr_symtab_b_create(struct wtr_vm* const vm)
{
	struct wtr_symtab_b* const symtab =
		(struct wtr_symtab_b*)wtr_vm_sys_malloc(vm,
		sizeof(struct wtr_symtab_b));

	if(NULL==symtab)
	{
		return NULL;
	}

	memset(symtab, 0, sizeof(*symtab));

	symtab->symtab.lookup  = wtr_symtab_b_lookup;
	symtab->symtab.destroy = wtr_symtab_b_destroy;

	wtr_list_init(&symtab->symtab.link_symtabs);

	if(0!=wtr_vector_create(&symtab->usersyms, vm, 8))
	{
		goto error;
	}

	/* register exceptions */
	{
		const struct wtr_sym_exc** p_exc = wtr_excsyms;

		while(NULL!=(*p_exc))
		{
			const struct wtr_sym_exc* const exc = *p_exc++;

			wt_globalid gid;

			if(WTE_OK!=wtr_gidtree_add(&vm->gidtree, vm, &gid,
				(struct wtr_sym*)exc) || gid!=exc->exc_id)
			{
				goto error;
			}
		}
	}

	wtr_list_insert_tail(&vm->list_symtabs, &symtab->symtab.link_symtabs);

	return symtab;

error:
	wtr_symtab_b_destroy((struct wtr_symtab*)symtab, vm);
	return NULL;
}

static struct wtr_sym* wtr_symtab_b_get_usertype(
	struct wtr_symtab* const symtab, wt_uint32_t usertype_id)
{
	/* resolution not supported */
	return NULL;
}

static struct wtr_sym* wtr_symtab_b_register(struct wtr_vm* const vm,
	struct wtr_symtab_b* const symtab, struct wtr_sym* const sym,
	wt_error* const out_err)
{
	if(WTE_OK!=((*out_err)=wtr_sym_bind(vm, sym)))
	{
		goto fail;
	}

	if(WTE_OK!=((*out_err)=wtr_sym_resolve(vm, sym,
		(struct wtr_symtab*)symtab, wtr_symtab_b_get_usertype)))
	{
		goto fail;
	}

	if(NULL!=wtr_vm_find_sym(vm, sym))
	{
		*out_err = WTE_AMBIG;
		goto fail;
	}

	if(0!=wtr_vector_push_back(&symtab->usersyms, vm, sym))
	{
		*out_err = WTE_NOMEM;
		goto fail;
	}

	return sym;

fail:;
	wtr_sym_unreg_gid(vm, sym);
	wtr_sym_release(vm, sym);
	return NULL;

}

struct wtr_sym_exc* wtr_symtab_b_add_exc(struct wtr_vm* const vm,
	struct wtr_symtab_b* const symtab, const char* const exc_name,
	wt_error* const out_err)
{
	const size_t namelen = strlen(exc_name) + 1;

	struct wtr_sym_exc* exc;

	if(WTR_MAX_SYMNAME_LEN<=namelen)
	{
		*out_err = WTE_SYMNAME;
		return NULL;
	}

	if(NULL==(exc=(struct wtr_sym_exc*)wtr_sym_new(vm, WTR_SYMTYPE_EXC,
		WTR_SYMBIND_EXPORT, (struct wtr_symtab*)symtab,
		sizeof(struct wtr_sym_exc), exc_name, namelen)))
	{
		*out_err = WTE_NOMEM;
		return NULL;
	}

	return (struct wtr_sym_exc*)wtr_symtab_b_register(vm, symtab,
		(struct wtr_sym*)exc, out_err);
}

struct wtr_sym_cfunc* wtr_symtab_b_add_func(struct wtr_vm* const vm,
	struct wtr_symtab_b* const symtab, const char* const func_name,
	wt_uint32_t argc, const wt_typeid* argv, wt_uint32_t retc,
	const wt_typeid* retv, WT_PFN_EXTFUNC func, void* token,
	wt_error* const out_err)
{
	const size_t namelen = strlen(func_name) + 1;

	struct wtr_sym_cfunc* cfunc;

	wt_typeid* typs;

	if(WTR_MAX_SYMNAME_LEN<=namelen)
	{
		*out_err = WTE_SYMNAME;
		return NULL;
	}

	if(NULL==(cfunc=(struct wtr_sym_cfunc*)wtr_sym_new(vm, WTR_SYMTYPE_CFUNC,
		WTR_SYMBIND_EXPORT, (struct wtr_symtab*)symtab,
		sizeof(struct wtr_sym_cfunc)+sizeof(wt_typeid)*(argc+retc),
		func_name, namelen)))
	{
		*out_err = WTE_NOMEM;
		return NULL;
	}

	typs = (wt_typeid*)(cfunc + 1);

	cfunc->cfunc            = func;
	cfunc->token            = token;
	cfunc->func.func_id     = WTR_SYM_UNDEF;
	cfunc->func.sig.argc    = argc;
	cfunc->func.sig.argv    = typs;
	cfunc->func.sig.retc    = retc;
	cfunc->func.sig.retv    = typs + argc;
	cfunc->func.sig.is_pure = 0;

	if(0<argc)
	{
		memcpy(typs, argv, sizeof(wt_typeid)*argc);
	}

	if(0<retc)
	{
		memcpy(typs+argc, retv, sizeof(wt_typeid)*retc);
	}

	return (struct wtr_sym_cfunc*)wtr_symtab_b_register(vm, symtab,
		(struct wtr_sym*)cfunc, out_err);
}
